/* classe che realizza la gestione degli accessi alla blackboard e la gestione
del passaggio da Java al motore Prolog dell'agente e viceversa */

import amzi.ls.*;
import java.util.Vector;
import java.rmi.*;
import java.rmi.registry.*;
import java.rmi.server.*;
import java.io.*;
import java.util.Date;

class logicAgent extends java.lang.Object
{
    LogicServer ls; /* motore Prolog */
    int instanceID; // istanza del logicAgent tra quelle attivate nell'ambito dell'agentServer
    agentServer as = null;  /* riferimento all'oggetto server dell'agente */
    private int readyThreadCount=0; // conta i thread che hanno terminato il run()

    public logicAgent( agentServer s ) throws LSException
    {
        as = s;
        instanceID = as.nextInstance();
        try {
            // create and initialize a logic server
            ls = new LogicServer();
            ls.Init( "" );

            // add extended predicates
            ls.AddPred( "ask_remote_cons", 3, "logicAgent", "askConsistency", this );
            ls.AddPred( "prompt", 1, "logicAgent", "prompt", this );
            ls.AddPred( "aggiorna", 1, "logicAgent", "aggiorna", this );
            // load the compiled Prolog expert system and the metainterpreter
            ls.Load( as.rules );
        } catch ( LSException e ) { System.out.println( "@@@ " + e.GetMsg() ); }
        as.printMessage( "Creato il logicAgent " + instanceID + " di " + as.my.name + "." );
    }

/* metodo che gestisce la derivazione abduttiva */
// la write sulla blackboard NON  gestita dal Prolog!!!
    public String solveAbducible( boolean priority, String goal ) throws LSException
    {
        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": solveAbducible( " + goal + " )");
        String result=null;
        try {
            // Caricamento metainterprete
            long s = ls.ExecStr("load([startabducible])");
            long term, risultato;
            String delta;
            // Accesso alla blackboard trasparente (In/Out == read/write)
            if( as.loggedIn )
                delta = as.b.read( priority );
            else
                // delta = new String( "[]" );
                delta = as.delta;
            // questo check di compatibilit  stato aggiuto in seguito e non  detto che sia corretto
            if( !h.compatible( delta, new String( "[" + goal + "]" ) ) ) {
                as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": " + goal + " is not compatible with " + delta );
                as.b.unlock();
            }
            else {
                // Passaggio dell'esecuzione al motore Prolog: invocazione query
                if( ( as.others==null )||( as.others.size()==0 )||( !as.loggedIn ) )
                    // derivazione locale
                    term = ls.ExecStr( "start_local_abductive_derivation([" + goal + "]," + delta + ",AbdOut)" );
                else
                    // derivazione distribuita
                    term = ls.ExecStr( "start_abductive_derivation([" + goal + "]," + delta + ",AbdOut)" );
                // Termine esecuzione Prolog: ritornata una risposta
                // true --> term>0, false --> term=0
                if( term != 0 ) {
                    risultato = ls.GetArg( term, 3 );   //risultato contiene la lista degli abducibili trovati
                    result = ls.TermToStr( risultato, 10000 );
                    if( ( as.loggedIn )&&( as.b.isAvailableForLogin() ) ) { // se non c' una procedura di login in atto
                        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": as.blackboard.write( " + result + " )" );
                        as.b.write( result );
                    }
                }
                else {
                    if( ( as.loggedIn )&&( as.b.isAvailableForLogin() ) ) { // se non c' una procedura di login in atto
                        as.b.unlock();
                        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": as.blackboard.unlock()" );
                    }
                }
            }
        } catch ( Exception e ) { e.printStackTrace(); }
        return result;
    }

    // versione con 'start_local_abductive_derivation'
    public String consistent( String goal ) throws LSException
    {
        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": checking the consistency of " + goal + " with ICs");
        String result=null, delta = new String( "[]" );
        try {
            // Caricamento metainterprete
            long s = ls.ExecStr("load([startabducible])");
            long term, risultato;
            // Passaggio dell'esecuzione al motore Prolog: invocazione query
            term = ls.ExecStr( "start_local_abductive_derivation([" + goal + "]," + delta + ",AbdOut)" );
            // viene anche eventualmente aggiornato il delta secondo gli IC dell'agente entrante
            // Termine esecuzione Prolog: ritornata una risposta
            // true --> term>0, false --> term=0
            if( term != 0 ) {
                risultato = ls.GetArg( term, 3 );   //risultato contiene la lista degli abducibili trovati
                result = ls.TermToStr( risultato, 10000 );
            }
        } catch ( Exception e ) { e.printStackTrace(); }
        return result;
    }
    
    // versione con 'set_of_local_abductive_derivation':
    // * voglio tutti i risultati, per poter fare backtracking
    // * in generale delta pu essere non vuoto ( quello che reca con s l'agente entrante)
    public Vector setof_consistent( String goal, String delta ) throws LSException
    {
        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": checking the consistency of " + goal + " with ICs");
        String result=null;
        try {
            // Caricamento metainterprete
            long s = ls.ExecStr("load([startabducible])");
            long term, risultato;
            // Passaggio dell'esecuzione al motore Prolog: invocazione query
            term = ls.ExecStr( "set_of_local_abductive_derivation([" + goal + "]," + delta + ",ListAbdOut)" );
            // viene anche eventualmente aggiornato il delta secondo gli IC dell'agente entrante
            // Termine esecuzione Prolog: ritornata una risposta
            // true --> term>0, false --> term=0
            if( term != 0 ) {
                risultato = ls.GetArg( term, 3 );   //risultato contiene la lista degli abducibili trovati
                result = ls.TermToStr( risultato, 10000 );
            }
            else { // fallita
                result = new String( "[]" );
            }
        } catch ( Exception e ) { e.printStackTrace(); }
        Vector v = h.stringToVector( result );
        return v;
    }
    
    /* metodo che gestisce la derivazione di consistenza */
    public Vector askConsLogicAgent( String goal, Vector abdListIn ) throws LSException
    {
        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": askConsLogicAgent( " + goal + ", " + abdListIn.toString() + " )" );
        try {
            /* Caricamento metainterprete */
            ls.ExecStr( "load([startabducible])" );
        } catch ( LSException e ) { System.out.println( "@@@ " + e.GetMsg() ); }
        long abdin, risultato, term, result;
        Vector v = new Vector();
        String ris = null, abdStr = null;
        /* Preparazione argomenti della richiesta di consistenza */
        abdin = ls.MakeList();
        for( int i=0; i<abdListIn.size(); i++ ) {
            term = ls.StrToTerm( ( String )abdListIn.elementAt( i ) );
            abdin = ls.PushList( abdin, term );
        } //abdin contiene la lista degli abducibili in ingresso
        /* Passaggio dell'esecuzione al motore Prolog: invocazione query */
        abdStr = ls.TermToStr( abdin, 10000 );
        term = ls.ExecStr( "solve_con(" + goal + "," + abdStr + ",AbdOut)" );
        /* Termine esecuzione Prolog: ritornata una risposta
        true --> term>0, false --> term=0 */
        if( term > 0 )
        {
            risultato = ls.GetArg( term, 3 );  //risultato contiene la lista degli abducibili trovati
            String risultatoStr = ls.TermToStr( risultato, 10000 );
            while ( risultato != 0 ) {
                result=ls.GetHead( risultato );
                ris = ls.TermToStr( result, 100 );
                v.addElement( ris );
                risultato = ls.GetTail( risultato );
            }
        // v rappresenta il vettore degli abducibili trovati
        // Se la query fallisce il vettore di ritorno v e' vuoto
        }
        return v;
    }

    // metodo richiamato dal motore Prolog quando viene incontrato il predicato
    // ask_remote_cons(Goal,AbdIn,AbdOut) che fa tornare il controllo a Java

/* ------------- non cancellare... ----------------

    // gestione sequenziale delle risposte dei thread

    public boolean askConsistency() throws LSException
    {
        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": askConsistency()" );
        Vector abducibili=new Vector();
        Vector abdOut=new Vector();
        String ThreadGroupName = new String( "ThreadGroup-" + as.ThreadGroupCount++ );
        ThreadGroup tg = new ThreadGroup( ThreadGroupName );
        String abd=null;
        long head;
        // Conversione argomenti Prolog in tipi Java
        String goal=ls.TermToStr( ls.GetParm(1), 100 );   //goal e' l'abducibile su cui verificare la consistency
        long lista=ls.GetParm(2); //lista contiene la lista degli abducibili trovati finora
        String listastr=ls.TermToStr( lista, 10000 );
        while ( lista!=0 ) {
            head=ls.GetHead( lista );
            abd=ls.TermToStr( head, 100 );
            abducibili.addElement( abd );
            lista=ls.GetTail( lista );
        }
        Vector waitThread = new Vector( as.others.size() );
        Vector risposta = new Vector();
        iAgent remoteServer;
        String serv = null;
        try {
            // Avvio di un thread per ogni agente del gruppo
            for( int i=0; i<as.others.size(); i++ ) {
                logicEnd le = ( logicEnd )as.others.elementAt( i );
                serv = "rmi://" + le.address + ":" + le.port + "/" + le.name;
                String n = new String( "Thread-" + ThreadGroupName.substring( 12, ThreadGroupName.length() ) + ":" + i );
                
                // ---- versione che usa askConsThread( String n, ThreadGroup tg, iAgent a, logicAgent l,  String g, Vector abd ) 
                askConsThread t = new askConsThread( n, tg, le.agent, this, goal, abducibili );
                t.start();
                waitThread.addElement( ( askConsThread )t );
                as.printMessage( "waitThread: " + waitThread );
                // ----------------



//                // ---- versione che usa askConsThread( String n, ThreadGroup tg, logicEnd le, logicAgent l,  String g, Vector abd ) 
//                askConsThread t = new askConsThread( n, tg, le, this, goal, abducibili );
//                t.start();
//                waitThread.addElement( ( askConsThread )t );
//               as.printMessage( "waitThread: " + waitThread );
//                // ----------------



//                // ---- versione che usa askConsThread( String n, ThreadGroup tg, iAgent a, logicAgent l,  String g, Vector abd ) 
//                try {
//                    remoteServer = ( iAgent )Naming.lookup( serv );
//                as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": invio richiesta di consistenza per il termine '" + goal + "' all'agente " + serv );
//                as.println( "askConsThread t = new askConsThread( " + n + ", " + tg + ", " + le + " [" + le.name + ", " + le.address + ", " + le.port + "], " + this + ", " + goal + ", " + abducibili + ");" );
//                as.printMessage( "askConsThread t = new askConsThread( " + n + ", " + tg + ", " + le + " [" + le.name + ", " + le.address + ", " + le.port + "], " + this + ", " + goal + ", " + abducibili + ");" );
//                    askConsThread t = new askConsThread( n, tg, remoteServer, this, goal, abducibili );
//                    t.start();
//                    waitThread.addElement( ( askConsThread )t );
//                    as.printMessage( "waitThread: " + waitThread );
//                } catch ( RemoteException ignored ) {
//                    System.out.println( "@@@ RemoteException in askConsThread(" + n + ") constructor at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
//                    as.printMessage( "@@@ RemoteException in askConsThread() constructor at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
//                } catch ( java.net.MalformedURLException ignored ) {
//                    System.out.println( "@@@ MalformedURLException in askConsThread(" + n + ") constructor at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
//                    as.printMessage( "@@@ MalformedURLException in askConsThread() constructor at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
//                } catch ( NotBoundException ignored ) {
//                    System.out.println( "@@@ NotBoundException in askConsThread(" + n + ") constructor at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
//                    as.printMessage( "@@@ NotBoundException in askConsThread() constructor at 'remoteServer = ( iAgent )Naming.lookup( " + serv + " );'" );
//                } // fine delle rmi
//                // ----------------
//

                }
            // Gestione sequenziale delle risposte ottenute da ciascun thread (l'ordine  fisso!!!!)
            for( int i=0; i<as.others.size(); i++ ) {
                askConsThread w = ( askConsThread )waitThread.firstElement();
                waitThread.removeElementAt( 0 );
                serv = w.getName();
                as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": attendo una risposta dal thread " + serv + "..." );
                w.join();
                risposta = w.getResult();
                if( risposta.size()>0 ) {
                    as.printMessage( "Ricevuta risposta per la consistenza del termine '" + goal + "' dal thread " + serv + "; ottenuta la lista risultato ", risposta );
                    for( int j=0; j<risposta.size(); j++ ) {
                        abdOut.addElement( risposta.elementAt(j) );
                    }
                } else {
                    as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": la richiesta di consistenza per il termine '" + goal + "' gestita da " + serv + " e' fallita" );
                    as.printMessage( ThreadGroupName + ".stop();" );
                    // ------
//                    for( int j=i+1; j<as.others.size(); j++ ) {
//                        w=( askConsThread )waitThread.firstElement();
//                        if( w!=null ) {
//                            waitThread.removeElementAt( 0 );
//                            if( w.isAlive() ) {
//                                w.stop();
//                            }
//                        }
//                    } // end for( int j=i+1; j<as.others.size(); j++ )
                    // ------
                    tg.stop(); // <--- versione precedente
                    return false;
                } // end else
            } // end for ( int i=0; i<as.others.size(); i++ )
        } catch ( InterruptedException e ) { 
            as.printMessage( "@@@ InterruptedException in askConsistency()" );
            e.printStackTrace(); } // fine del thread
        long termList = ls.MakeList();
        long termine;
        for( int i=0; i<abdOut.size(); i++ ) {
            termine = ls.StrToTerm( ( String )abdOut.elementAt( i ) );
            termList = ls.PushList( termList, termine );
        }
        ls.UnifyParm( 3, termList );
        if( tg==null ) System.out.println( "@@@ " + ThreadGroupName + " is null!!!" );
        return true;
    }
    // -------------- ... fin qui -----------------
*/


    // gestione non sequenziale delle risposte (utilizza i segnali)
    // bisogna ancora verificare se effettivamente funziona.
    
    public boolean askConsistency() throws LSException
    {
        // la consistenza locale viene gestita a monte (uso di start_local_abductive_derivation/3)
        // per cui la parte sotto commento che segue  inutile.
//        // consento di operare in locale, se la lista degli altri agenti  vuota
//        if( ( as.others==null )||( as.others.size()==0 ) )
//        {
//            as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": askConsistency( local )" );
//            Vector abducibili=new Vector();
//            String goal=ls.TermToStr( ls.GetParm(1), 100 );   //goal e' l'abducibile su cui verificare la consistency
//            long head, lista=ls.GetParm(2); //lista contiene la lista degli abducibili trovati finora
//            String listastr=ls.TermToStr( lista, 10000 ), abd=null;
//            while ( lista!=0 ) {
//                head = ls.GetHead( lista );
//                abd = ls.TermToStr( head, 100 );
//                abducibili.addElement( abd );
//                lista = ls.GetTail( lista );
//            }
//            long termine, termList = ls.MakeList();
//            for( int i=0; i<abducibili.size(); i++ ) {
//                termine = ls.StrToTerm( ( String )abducibili.elementAt( i ) );
//                termList = ls.PushList( termList, termine );
//            }
//            ls.UnifyParm( 3, termList );
//            return true;
//        }
        // altrimenti faccio partire le varie richieste di consistenza
        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": askConsistency( remote )" );
        Vector abducibili=new Vector();
        Vector abdOut=new Vector();
        String ThreadGroupName = new String( "ThreadGroup-" + as.ThreadGroupCount++ );
        ThreadGroup tg = new ThreadGroup( ThreadGroupName );
        String abd=null;
        long head;
        // Conversione argomenti Prolog in tipi Java
        String goal=ls.TermToStr( ls.GetParm(1), 100 );   //goal e' l'abducibile su cui verificare la consistency
        long lista=ls.GetParm(2); //lista contiene la lista degli abducibili trovati finora
        String listastr=ls.TermToStr( lista, 10000 );
        while ( lista!=0 ) {
            head=ls.GetHead( lista );
            abd=ls.TermToStr( head, 100 );
            abducibili.addElement( abd );
            lista=ls.GetTail( lista );
        }
        Vector waitThread = new Vector( as.others.size() );
        Vector risposta = new Vector();
        String serv = null;
        try {
            // Avvio di un thread per ogni agente del gruppo
            for( int i=0; i<as.others.size(); i++ ) {
                logicEnd le = ( logicEnd )as.others.elementAt( i );
                serv = "rmi://" + le.address + ":" + le.port + "/" + le.name;
                String n = new String( "Thread-" + ThreadGroupName.substring( 12, ThreadGroupName.length() ) + ":" + i );
                askConsThread t = new askConsThread( n, tg, le.agent, this, goal, abducibili );
                t.start();
                waitThread.addElement( ( askConsThread )t );
                as.printMessage( "waitThread: " + waitThread );
            }
            int index = 0;
            while( waitThread.size()>0 )
            {
                as.printMessage( "index = " + index );
                waitReadyThreadCount();
                askConsThread w = ( askConsThread )waitThread.elementAt( index );
                serv = w.getName();
                if( w.hasFinished() ) {
                    w.join();
                    risposta = w.getResult();
                    waitThread.removeElementAt( index );
                    decReadyThreadCount();
                    if( risposta.size()>0 ) {
                        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": ricevuta risposta per la consistenza del termine '" + goal + "' dal thread " + serv + "; ottenuta la lista risultato ", risposta );
                        for( int j=0; j<risposta.size(); j++ ) {
                            abdOut.addElement( risposta.elementAt(j) );
                        }
                    } else {
                        as.printMessage( "logicAgent " + instanceID + " di " + as.my.name + ": la richiesta di consistenza per il termine '" + goal + "' gestita da " + serv + " e' fallita" );
                        tg.stop();
                        return false;
                    } // end else( risposta.size()>0 )
                } // end if( w.hasFinished() )
                else {
                    index++;
                } // end else( w.hasFinished() )
                if( index >= waitThread.size() ) index=0;
            } // end while( waitThread.size()>0 )
        } catch ( Exception e ) { e.printStackTrace( as.outStream ); } // fine del thread
        long termList = ls.MakeList();
        long termine;
        for( int i=0; i<abdOut.size(); i++ ) {
            termine = ls.StrToTerm( ( String )abdOut.elementAt( i ) );
            termList = ls.PushList( termList, termine );
        }
        ls.UnifyParm( 3, termList );
        return true;
    }

public synchronized void waitReadyThreadCount() {
    if( readyThreadCount == 0 ) {
        try {
            wait();
        } catch ( InterruptedException e ) {
            e.printStackTrace(); }
    }
}

/* metodo richiamato dal motore Prolog quando viene incontrato il predicato prompt/1 */
    public synchronized boolean prompt() throws LSException
    {
        as.printMessage( "PROLOG (istanza " + instanceID + " di " + as.my.name + "): " + ls.TermToStr( ls.GetParm(1),10000 ) );
        return true;
    }

/* metodo richiamato dal motore Prolog quando viene incontrato il predicato aggiorna/1
per l'aggiornamento della blackboard nel caso di gestione tramite la politica read/write */
    public synchronized boolean aggiorna() throws LSException
    {
        boolean res = false;
        long lista = ls.GetParm(1);  //lista contiene la lista degli abducibili trovati
        String listastr = ls.TermToStr( lista, 10000 );
        return write( listastr );
    }
    
    private synchronized boolean write( String delta ) {
        boolean res = false;
        try {
            res = as.b.write( delta );
            if ( !res ) {
                as.printMessage( "Il risultato trovato e' in conflitto con quello presente nella blackboard" );
                as.printMessage( "La write non ha avuto successo" );
            } else { as.printMessage( "La write ha avuto successo" ); }
        } catch ( RemoteException e ) { System.out.println( "@@@ Errore in fase di aggiornamento della blackboard" ); }
        return res;
    }

    public synchronized void incReadyThreadCount() {
        readyThreadCount++; notifyAll();
        as.printMessage( "readyThreadCount = " + readyThreadCount );
    }

    public synchronized void decReadyThreadCount() {
        readyThreadCount--;
        if( readyThreadCount<0 ) {
            as.printMessage( "istanza " + instanceID + " di " + as.my.name + "): " + "!!!!Errore nella gestione dei thread (readyThreadCount<0)!!!!" );
            readyThreadCount=0;
        }
        as.printMessage( "readyThreadCount = " + readyThreadCount );
    }

    public void close() throws LSException
    {
        if (ls != null)
        ls.Close();
    }

}